from .simple_losses import CrossEntropyLoss, MultiObjectiveLoss
from .classificaion import DifferentiableArgmax, ParabolicPenaltyLoss, TailedCategoricalCrossentropyLoss
from .probabilistic import HeteroscedasticityLoss, DistributionNLL
from .soft_dtw import SoftDTW